/*==============================================================================*/
/*  Copyright (C) 2020, Heartland Data inc. All Rights Reserved.                */
/*                                                                              */
/*  Title  :   Ethernet Driver                                                  */
/*  EventID:   Via the driver                                                   */
/*  AppVer :   DT+ Ver1.00`                                                    */
/*  FileID :   21c                                                              */
/*  Version:   1.0                                                              */
/*  Author :   HLDC                                                             */
/*==============================================================================*/

import java.io.IOException;
import java.net.*;
import java.util.concurrent.TimeUnit;

public class DtTestPointDriver {
    
    /*--------------------------------------------------------------------------*/
    /*  DynamicTracar MaskCode                                                  */
    /*  Please do not change the value.                                         */
    /*--------------------------------------------------------------------------*/
    static public final int DT_VARIABLE_BIT         =   0x02;               /* Memory Dump MASK */
    static public final int DT_EVTTRG_BIT           =   0x08;               /* EventTrigger MASK */
    /*--------------------------------------------------------------------------*/
    /*  value Variables length. never set the value over 256                    */
    /*  Please do not change the value.                                         */
    /*--------------------------------------------------------------------------*/
    static public final int DT_MAX_VAL_SIZE         =   256;
    
    /*--------------------------------------------------------------------------*/
    /*  Please set 1 when you add Event Value to Test Point                     */
    /*--------------------------------------------------------------------------*/
    static public final int DT_ADD_EVENT_INFO       =   1;
    /*--------------------------------------------------------------------------*/
    /*  if DT_ADD_EVENT_INFO is 1. Please set 4.                                */
    /*--------------------------------------------------------------------------*/
    static public final int DT_EVENT_INFO_SIZE      =   4;
    /* if DT_ADD_EVENT_INFO is 1. set event data TP_BusOut and DtTestPointValue.*/

    /*--------------------------------------------------------------------------*/
    /*  Please set 1 when you add Time Value to Test Point                      */
    /*--------------------------------------------------------------------------*/
    static public final int DT_ADD_TIME_INFO        =   0;
    /*--------------------------------------------------------------------------*/
    /*  if DT_ADD_TIME_INFO is 1. Please set 8.                                 */
    /*--------------------------------------------------------------------------*/
    static public final int DT_TIME_INFO_SIZE       =   8;
    /*  if DT_ADD_TIME_INFO is 1. set time data TP_BusOut and DtTestPointValue. */

    /*--------------------------------------------------------------------------*/
    /*  DT_ADD_EVENT_INFO == 1 & DT_ADD_TIME_INFO == 1                          */
    /*                     -> Normal TP with Event & Time info. Please set 0x15.*/
    /*  DT_ADD_EVENT_INFO == 1 & DT_ADD_TIME_INFO == 0                          */
    /*                     -> Normal TP with Event info. Please set 0x05.       */
    /*  DT_ADD_EVENT_INFO == 0 & DT_ADD_TIME_INFO == 1                          */
    /*                     -> Normal TP with Time info. Please set 0x11.        */
    /*  DT_ADD_EVENT_INFO == 0 & DT_ADD_TIME_INFO == 0                          */
    /*                     -> Normal TP. Please set 0x01.                       */
    /*--------------------------------------------------------------------------*/
    static public final int DT_NRMLTP_ID            =   0x05;
    /*--------------------------------------------------------------------------*/
    /*  DT_ADD_EVENT_INFO == 1 & DT_ADD_TIME_INFO == 1                          */
    /*                   -> Variable TP with Event & Time info. Please set 0x16.*/
    /*  DT_ADD_EVENT_INFO == 1 & DT_ADD_TIME_INFO == 0                          */
    /*                   -> Variable TP with Event info. Please set 0x06.       */
    /*  DT_ADD_EVENT_INFO == 0 & DT_ADD_TIME_INFO == 1                          */
    /*                   -> Variable TP with Time info. Please set 0x12.        */
    /*  DT_ADD_EVENT_INFO == 0 & DT_ADD_TIME_INFO == 0                          */
    /*                   -> Variable TP. Please set 0x02.                       */
    /*--------------------------------------------------------------------------*/
    static public final int DT_VALTP_ID             =   0x06;

    /*--------------------------------------------------------------------------*/
    /*  DynamicTracar TP size info.                                             */
    /*  Please do not change the value.                                         */
    /*--------------------------------------------------------------------------*/
    static public final int DT_HEADER_SIZE          =   4;
    static public final int DT_NRMLTP_SIZE          =   DT_HEADER_SIZE + 6 + DT_EVENT_INFO_SIZE + DT_TIME_INFO_SIZE;
    static public final int DT_MAX_VALTP_SIZE       =   DT_NRMLTP_SIZE + 1 + DT_MAX_VAL_SIZE;
    static public final int DT_EVENTTP_SIZE         =   DT_NRMLTP_SIZE;

    static public String DT_IP_ADDRESS = "192.168.0.4"; /* Customize */
    static public int DT_PORT = 3000;
    static public DatagramSocket socket = null;
    static public InetAddress serverInet = null;
    static public DatagramPacket packet = null;
    static public DatagramPacket val_packet = null;
    static public byte[] buff = new byte[DT_NRMLTP_SIZE];
    static public byte[] val_buff = new byte[DT_MAX_VALTP_SIZE];
    static long start_time = 0;
    static long time = 0;
    static int event = 0;

    /*--------------------------------------------------------------------------*/
    /*  TestPoint InsertCode Method                                             */
    /*--------------------------------------------------------------------------*/
    /* Normal TestPoint Method */
    static public final void DtTestPoint(int nBaseAddress, int nStepNumber){
        TpBusOut( nBaseAddress, nStepNumber);
    }

    /* Performance TestPoint Method */
    static public final void DtTestPointIdle(int nIdleAddress, int nStepNumber){
        /* Call DynamicTracerDriver */
        TpBusOut(nIdleAddress, nStepNumber);
    }

    /* EventTrigger TestPoint Method */
    static public final void DtTestPointEventTrigger(int nBaseAddress, int nStepNumber, int theValue){
        /* Call DynamicTracerDriver */
        int count = 0;
        int packet_size = DT_NRMLTP_SIZE + 2;

        val_buff[count++] = (byte)0xFF;                     /* Header */
        val_buff[count++] = (byte)0xFF;                     /* Header */
        val_buff[count++] = (byte)(packet_size);            /* Data Size */
        val_buff[count++] = (byte)((packet_size) >> 8);     /* Data Size */
        val_buff[count++] = (byte)DT_VALTP_ID;              /* Variable TP */
        val_buff[count++] = (byte)(nStepNumber);            /* nStepNumber */
        val_buff[count++] = (byte)(nStepNumber >> 8);       /* nStepNumber */
        val_buff[count++] = (byte)(nBaseAddress | DT_EVTTRG_BIT);    /* nBaseAddress */
        val_buff[count++] = (byte)(nBaseAddress >> 8);      /* nBaseAddress */
        val_buff[count++] = (byte)(nBaseAddress >> 16);     /* nBaseAddress */
        if(DT_ADD_TIME_INFO == 1){
            DtTimeSetInfo(count);
            count = count + DT_TIME_INFO_SIZE;
        }
        val_buff[count++] = (byte)theValue;                 /* value */
        val_buff[count++] = (byte)(theValue >> 8);          /* value */

        TP_MemoryOutput(packet_size);
    }

    /* EventTrigger32 TestPoint Method */
    static public final void DtTestPointEventTrigger32(int nBaseAddress, int nStepNumber, int theValue){
        /* Call DynamicTracerDriver */
        int count = 0;
        int packet_size = DT_NRMLTP_SIZE + 4;

        val_buff[count++] = (byte)0xFF;                     /* Header */
        val_buff[count++] = (byte)0xFF;                     /* Header */
        val_buff[count++] = (byte)(packet_size);            /* Data Size */
        val_buff[count++] = (byte)((packet_size) >> 8);     /* Data Size */
        val_buff[count++] = (byte)DT_VALTP_ID;              /* Variable TP */
        val_buff[count++] = (byte)(nStepNumber);            /* nStepNumber */
        val_buff[count++] = (byte)(nStepNumber >> 8);       /* nStepNumber */
        val_buff[count++] = (byte)(nBaseAddress | DT_EVTTRG_BIT);    /* nBaseAddress */
        val_buff[count++] = (byte)(nBaseAddress >> 8);      /* nBaseAddress */
        val_buff[count++] = (byte)(nBaseAddress >> 16);     /* nBaseAddress */
        if(DT_ADD_TIME_INFO == 1){
            DtTimeSetInfo(count);
            count = count + DT_TIME_INFO_SIZE;
        }
        val_buff[count++] = (byte)theValue;                 /* value */
        val_buff[count++] = (byte)(theValue >> 8);          /* value */
        val_buff[count++] = (byte)(theValue >> 16);         /* value */
        val_buff[count++] = (byte)(theValue >> 24);         /* value */

        TP_MemoryOutput(packet_size);
    }

    /* Set Time Info */
    static public final void DtTimeSetInfo(int count){
        time = TimeUnit.MICROSECONDS.convert( System.nanoTime() - start_time, TimeUnit.NANOSECONDS );
        val_buff[count++] = (byte)(time);
        val_buff[count++] = (byte)(time >> 8);
        val_buff[count++] = (byte)(time >> 16);
        val_buff[count++] = (byte)(time >> 24);
        val_buff[count++] = (byte)(time >> 32);
        val_buff[count++] = (byte)(time >> 40);
        val_buff[count++] = (byte)(time >> 48);
        val_buff[count++] = (byte)(time >> 56);
    }

    /* Set Event Info */
    static public final void DtEventSetInfo(int count){
        event = (int)Thread.currentThread().getId();
        val_buff[count++] = (byte)(event);
        val_buff[count++] = (byte)(event >> 8);
        val_buff[count++] = (byte)(event >> 16);
        val_buff[count++] = (byte)(event >> 24);
    }

    /* MemoryOutput TestPoint Method */
    static public final void TP_MemoryOutput(int packet_size){
        if (socket == null) {
            TpBusEthernetInit();
            if(socket == null){
                return;
            }
        }
        try{
            val_packet.setLength(packet_size);
            socket.send(val_packet);
        }
        catch(IOException e){
        }
    }

    /* MemoryOutput TestPoint Method */
    static public final void DtTestPointValue(int nBaseAddress, int nStepNumber, byte theValue, int nLength){
        int count = 0;
        int packet_size = DT_NRMLTP_SIZE + 1 + 1;

        val_buff[count++] = (byte)0xFF;                     /* Header */
        val_buff[count++] = (byte)0xFF;                     /* Header */
        val_buff[count++] = (byte)(packet_size);            /* Data Size */
        val_buff[count++] = (byte)((packet_size) >> 8);     /* Data Size */
        val_buff[count++] = (byte)DT_VALTP_ID;              /* Variable TP */
        val_buff[count++] = (byte)(nStepNumber);            /* nStepNumber */
        val_buff[count++] = (byte)(nStepNumber >> 8);       /* nStepNumber */
        val_buff[count++] = (byte)(nBaseAddress | DT_VARIABLE_BIT);    /* nBaseAddress */
        val_buff[count++] = (byte)(nBaseAddress >> 8);      /* nBaseAddress */
        val_buff[count++] = (byte)(nBaseAddress >> 16);     /* nBaseAddress */
        if(DT_ADD_TIME_INFO == 1){
            DtTimeSetInfo(count);
            count = count + DT_TIME_INFO_SIZE;
        }
        if(DT_ADD_EVENT_INFO == 1){
            DtEventSetInfo(count);
            count = count + DT_EVENT_INFO_SIZE;
        }
        val_buff[count++] = (byte)1;                        /* size */
        val_buff[count++] = (byte)theValue;                 /* value */

        TP_MemoryOutput(packet_size);
    }
    static public final void DtTestPointValue(int nBaseAddress, int nStepNumber, char theValue, int nLength){
        int count = 0;
        int packet_size = DT_NRMLTP_SIZE + 1 + 2;

        val_buff[count++] = (byte)0xFF;                     /* Header */
        val_buff[count++] = (byte)0xFF;                     /* Header */
        val_buff[count++] = (byte)(packet_size);            /* Data Size */
        val_buff[count++] = (byte)((packet_size) >> 8);     /* Data Size */
        val_buff[count++] = (byte)DT_VALTP_ID;              /* Variable TP */
        val_buff[count++] = (byte)(nStepNumber);            /* nStepNumber */
        val_buff[count++] = (byte)(nStepNumber >> 8);       /* nStepNumber */
        val_buff[count++] = (byte)(nBaseAddress | DT_VARIABLE_BIT);    /* nBaseAddress */
        val_buff[count++] = (byte)(nBaseAddress >> 8);      /* nBaseAddress */
        val_buff[count++] = (byte)(nBaseAddress >> 16);     /* nBaseAddress */
        if(DT_ADD_TIME_INFO == 1){
            DtTimeSetInfo(count);
            count = count + DT_TIME_INFO_SIZE;
        }
        if(DT_ADD_EVENT_INFO == 1){
            DtEventSetInfo(count);
            count = count + DT_EVENT_INFO_SIZE;
        }
        val_buff[count++] = (byte)2;                        /* size */
        val_buff[count++] = (byte)theValue;                 /* value */
        val_buff[count++] = (byte)(theValue >> 8);          /* value */
        TP_MemoryOutput(packet_size);
    }
    static public final void DtTestPointValue(int nBaseAddress, int nStepNumber, short theValue, int nLength){
        int count = 0;
        int packet_size = DT_NRMLTP_SIZE + 1 + 2;

        val_buff[count++] = (byte)0xFF;                     /* Header */
        val_buff[count++] = (byte)0xFF;                     /* Header */
        val_buff[count++] = (byte)(packet_size);            /* Data Size */
        val_buff[count++] = (byte)((packet_size) >> 8);     /* Data Size */
        val_buff[count++] = (byte)DT_VALTP_ID;              /* Variable TP */
        val_buff[count++] = (byte)(nStepNumber);            /* nStepNumber */
        val_buff[count++] = (byte)(nStepNumber >> 8);       /* nStepNumber */
        val_buff[count++] = (byte)(nBaseAddress | DT_VARIABLE_BIT);    /* nBaseAddress */
        val_buff[count++] = (byte)(nBaseAddress >> 8);      /* nBaseAddress */
        val_buff[count++] = (byte)(nBaseAddress >> 16);     /* nBaseAddress */
        if(DT_ADD_TIME_INFO == 1){
            DtTimeSetInfo(count);
            count = count + DT_TIME_INFO_SIZE;
        }
        if(DT_ADD_EVENT_INFO == 1){
            DtEventSetInfo(count);
            count = count + DT_EVENT_INFO_SIZE;
        }
        val_buff[count++] = (byte)2;                        /* size */
        val_buff[count++] = (byte)theValue;                 /* value */
        val_buff[count++] = (byte)(theValue >> 8);          /* value */
        TP_MemoryOutput(packet_size);
    }
    static public final void DtTestPointValue(int nBaseAddress, int nStepNumber, int theValue, int nLength){
        int count = 0;
        int packet_size = DT_NRMLTP_SIZE + 1 + 4;

        val_buff[count++] = (byte)0xFF;                     /* Header */
        val_buff[count++] = (byte)0xFF;                     /* Header */
        val_buff[count++] = (byte)(packet_size);            /* Data Size */
        val_buff[count++] = (byte)((packet_size) >> 8);     /* Data Size */
        val_buff[count++] = (byte)DT_VALTP_ID;              /* Variable TP */
        val_buff[count++] = (byte)(nStepNumber);            /* nStepNumber */
        val_buff[count++] = (byte)(nStepNumber >> 8);       /* nStepNumber */
        val_buff[count++] = (byte)(nBaseAddress | DT_VARIABLE_BIT);    /* nBaseAddress */
        val_buff[count++] = (byte)(nBaseAddress >> 8);      /* nBaseAddress */
        val_buff[count++] = (byte)(nBaseAddress >> 16);     /* nBaseAddress */
        if(DT_ADD_TIME_INFO == 1){
            DtTimeSetInfo(count);
            count = count + DT_TIME_INFO_SIZE;
        }
        if(DT_ADD_EVENT_INFO == 1){
            DtEventSetInfo(count);
            count = count + DT_EVENT_INFO_SIZE;
        }
        val_buff[count++] = (byte)4;                        /* size */
        val_buff[count++] = (byte)theValue;                 /* value */
        val_buff[count++] = (byte)(theValue >> 8);          /* value */
        val_buff[count++] = (byte)(theValue >> 16);         /* value */
        val_buff[count++] = (byte)(theValue >> 24);         /* value */
        TP_MemoryOutput(packet_size);
    }
    static public final void DtTestPointValue(int nBaseAddress, int nStepNumber, long theValue, int nLength){
        int count = 0;
        int packet_size = DT_NRMLTP_SIZE + 1 + 8;

        val_buff[count++] = (byte)0xFF;                     /* Header */
        val_buff[count++] = (byte)0xFF;                     /* Header */
        val_buff[count++] = (byte)(packet_size);            /* Data Size */
        val_buff[count++] = (byte)((packet_size) >> 8);     /* Data Size */
        val_buff[count++] = (byte)DT_VALTP_ID;              /* Variable TP */
        val_buff[count++] = (byte)(nStepNumber);            /* nStepNumber */
        val_buff[count++] = (byte)(nStepNumber >> 8);       /* nStepNumber */
        val_buff[count++] = (byte)(nBaseAddress | DT_VARIABLE_BIT);    /* nBaseAddress */
        val_buff[count++] = (byte)(nBaseAddress >> 8);      /* nBaseAddress */
        val_buff[count++] = (byte)(nBaseAddress >> 16);     /* nBaseAddress */
        if(DT_ADD_TIME_INFO == 1){
            DtTimeSetInfo(count);
            count = count + DT_TIME_INFO_SIZE;
        }
        if(DT_ADD_EVENT_INFO == 1){
            DtEventSetInfo(count);
            count = count + DT_EVENT_INFO_SIZE;
        }
        val_buff[count++] = (byte)8;                        /* size */
        val_buff[count++] = (byte)theValue;                 /* value */
        val_buff[count++] = (byte)(theValue >> 8);          /* value */
        val_buff[count++] = (byte)(theValue >> 16);         /* value */
        val_buff[count++] = (byte)(theValue >> 24);         /* value */
        val_buff[count++] = (byte)(theValue >> 32);         /* value */
        val_buff[count++] = (byte)(theValue >> 40);         /* value */
        val_buff[count++] = (byte)(theValue >> 48);         /* value */
        val_buff[count++] = (byte)(theValue >> 56);         /* value */
        TP_MemoryOutput(packet_size);
    }
    static public final void DtTestPointValue(int nBaseAddress, int nStepNumber, boolean theValue, int nLength){
        int count = 0;
        int packet_size = DT_NRMLTP_SIZE + 1 + 1;

        val_buff[count++] = (byte)0xFF;                     /* Header */
        val_buff[count++] = (byte)0xFF;                     /* Header */
        val_buff[count++] = (byte)(packet_size);            /* Data Size */
        val_buff[count++] = (byte)((packet_size) >> 8);     /* Data Size */
        val_buff[count++] = (byte)DT_VALTP_ID;              /* Variable TP */
        val_buff[count++] = (byte)(nStepNumber);            /* nStepNumber */
        val_buff[count++] = (byte)(nStepNumber >> 8);       /* nStepNumber */
        val_buff[count++] = (byte)(nBaseAddress | DT_VARIABLE_BIT);    /* nBaseAddress */
        val_buff[count++] = (byte)(nBaseAddress >> 8);      /* nBaseAddress */
        val_buff[count++] = (byte)(nBaseAddress >> 16);     /* nBaseAddress */
        if(DT_ADD_TIME_INFO == 1){
            DtTimeSetInfo(count);
            count = count + DT_TIME_INFO_SIZE;
        }
        if(DT_ADD_EVENT_INFO ==1){
            DtEventSetInfo(count);
            count = count + DT_EVENT_INFO_SIZE;
        }
        val_buff[count++] = (byte)1;                    /* size */
        val_buff[count++] = (byte)(theValue?0x01:0x00); /* value */
        TP_MemoryOutput(packet_size);
    }
    static public final void DtTestPointValue(int nBaseAddress, int nStepNumber, byte[] theValue, int nLength){
        if( nLength > 0 ){
            int count = 0;
            int size = ((1 * nLength) < DT_MAX_VAL_SIZE) ? (1 * nLength) : DT_MAX_VAL_SIZE;
            int packet_size = DT_NRMLTP_SIZE + 1 + size;

            val_buff[count++] = (byte)0xFF;                     /* Header */
            val_buff[count++] = (byte)0xFF;                     /* Header */
            val_buff[count++] = (byte)(packet_size);            /* Data Size */
            val_buff[count++] = (byte)((packet_size) >> 8);     /* Data Size */
            val_buff[count++] = (byte)DT_VALTP_ID;              /* Variable TP */
            val_buff[count++] = (byte)(nStepNumber);            /* nStepNumber */
            val_buff[count++] = (byte)(nStepNumber >> 8);       /* nStepNumber */
            val_buff[count++] = (byte)(nBaseAddress | DT_VARIABLE_BIT);    /* nBaseAddress */
            val_buff[count++] = (byte)(nBaseAddress >> 8);      /* nBaseAddress */
            val_buff[count++] = (byte)(nBaseAddress >> 16);     /* nBaseAddress */
            if(DT_ADD_TIME_INFO == 1){
                DtTimeSetInfo(count);
                count = count + DT_TIME_INFO_SIZE;
            }
            if(DT_ADD_EVENT_INFO == 1){
                DtEventSetInfo(count);
                count = count + DT_EVENT_INFO_SIZE;
            }
            val_buff[count++] = (byte)(size);                   /* size */
            for(int cnt = 0; cnt < nLength && count < packet_size; cnt++){
                val_buff[count++] = (byte)theValue[cnt];        /* value */
            }
            TP_MemoryOutput(packet_size);
        }
    }
    static public final void DtTestPointValue(int nBaseAddress, int nStepNumber, char[] theValue, int nLength){
        if( nLength > 0 ){
            int count = 0;
            int size = ((2 * nLength) < DT_MAX_VAL_SIZE) ? (2 * nLength) : DT_MAX_VAL_SIZE;
            int packet_size = DT_NRMLTP_SIZE + 1 + size;

            val_buff[count++] = (byte)0xFF;                         /* Header */
            val_buff[count++] = (byte)0xFF;                         /* Header */
            val_buff[count++] = (byte)(packet_size);                /* Data Size */
            val_buff[count++] = (byte)((packet_size) >> 8);         /* Data Size */
            val_buff[count++] = (byte)DT_VALTP_ID;                  /* Variable TP */
            val_buff[count++] = (byte)(nStepNumber);                /* nStepNumber */
            val_buff[count++] = (byte)(nStepNumber >> 8);           /* nStepNumber */
            val_buff[count++] = (byte)(nBaseAddress | DT_VARIABLE_BIT);        /* nBaseAddress */
            val_buff[count++] = (byte)(nBaseAddress >> 8);          /* nBaseAddress */
            val_buff[count++] = (byte)(nBaseAddress >> 16);         /* nBaseAddress */
            if(DT_ADD_TIME_INFO == 1){
                DtTimeSetInfo(count);
                count = count + DT_TIME_INFO_SIZE;
            }
            if(DT_ADD_EVENT_INFO == 1){
                DtEventSetInfo(count);
                count = count + DT_EVENT_INFO_SIZE;
            }
            val_buff[count++] = (byte)(size);                       /* size */
            for(int cnt = 0; cnt < nLength && count < packet_size; cnt++){
                val_buff[count++] = (byte)theValue[cnt];            /* value */
                val_buff[count++] = (byte)(theValue[cnt] >> 8);     /* value */
            }
            TP_MemoryOutput(packet_size);
        }
    }
    static public final void DtTestPointValue(int nBaseAddress, int nStepNumber, short[] theValue, int nLength){
        if( nLength > 0 ){
            int count = 0;
            int size = ((2 * nLength) < DT_MAX_VAL_SIZE) ? (2 * nLength) : DT_MAX_VAL_SIZE;
            int packet_size = DT_NRMLTP_SIZE + 1 + size;

            val_buff[count++] = (byte)0xFF;                         /* Header */
            val_buff[count++] = (byte)0xFF;                         /* Header */
            val_buff[count++] = (byte)(packet_size);                /* Data Size */
            val_buff[count++] = (byte)((packet_size) >> 8);         /* Data Size */
            val_buff[count++] = (byte)DT_VALTP_ID;                  /* Variable TP */
            val_buff[count++] = (byte)(nStepNumber);                /* nStepNumber */
            val_buff[count++] = (byte)(nStepNumber >> 8);           /* nStepNumber */
            val_buff[count++] = (byte)(nBaseAddress | DT_VARIABLE_BIT);        /* nBaseAddress */
            val_buff[count++] = (byte)(nBaseAddress >> 8);          /* nBaseAddress */
            val_buff[count++] = (byte)(nBaseAddress >> 16);         /* nBaseAddress */
            if(DT_ADD_TIME_INFO == 1){
                DtTimeSetInfo(count);
                count = count + DT_TIME_INFO_SIZE;
            }
            if(DT_ADD_EVENT_INFO == 1){
                DtEventSetInfo(count);
                count = count + DT_EVENT_INFO_SIZE;
            }
            val_buff[count++] = (byte)(size);                       /* size */
            for(int cnt = 0; cnt < nLength && count < packet_size; cnt++){
                val_buff[count++] = (byte)theValue[cnt];            /* value */
                val_buff[count++] = (byte)(theValue[cnt] >> 8);     /* value */
            }
            TP_MemoryOutput(packet_size);
        }
    }
    static public final void DtTestPointValue(int nBaseAddress, int nStepNumber, int[] theValue, int nLength){
        if( nLength > 0 ){
            int count = 0;
            int size = ((4 * nLength) < DT_MAX_VAL_SIZE) ? (4 * nLength) : DT_MAX_VAL_SIZE;
            int packet_size = DT_NRMLTP_SIZE + 1 + size;

            val_buff[count++] = (byte)0xFF;                         /* Header */
            val_buff[count++] = (byte)0xFF;                         /* Header */
            val_buff[count++] = (byte)(packet_size);                /* Data Size */
            val_buff[count++] = (byte)((packet_size) >> 8);         /* Data Size */
            val_buff[count++] = (byte)DT_VALTP_ID;                  /* Variable TP */
            val_buff[count++] = (byte)(nStepNumber);                /* nStepNumber */
            val_buff[count++] = (byte)(nStepNumber >> 8);           /* nStepNumber */
            val_buff[count++] = (byte)(nBaseAddress | DT_VARIABLE_BIT);        /* nBaseAddress */
            val_buff[count++] = (byte)(nBaseAddress >> 8);          /* nBaseAddress */
            val_buff[count++] = (byte)(nBaseAddress >> 16);         /* nBaseAddress */
            if(DT_ADD_TIME_INFO ==1){
                DtTimeSetInfo(count);
                count = count + DT_TIME_INFO_SIZE;
            }
            if(DT_ADD_EVENT_INFO == 1){
                DtEventSetInfo(count);
                count = count + DT_EVENT_INFO_SIZE;
            }
            val_buff[count++] = (byte)(size);                       /* size */
            for(int cnt = 0; cnt < nLength && count < packet_size; cnt++){
                val_buff[count++] = (byte)theValue[cnt];            /* value */
                val_buff[count++] = (byte)(theValue[cnt] >> 8);     /* value */
                val_buff[count++] = (byte)(theValue[cnt] >> 16);    /* value */
                val_buff[count++] = (byte)(theValue[cnt] >> 24);    /* value */
            }
            TP_MemoryOutput(packet_size);
        }
    }
    static public final void DtTestPointValue(int nBaseAddress, int nStepNumber, long[] theValue, int nLength){
        if( nLength > 0 ){
            int count = 0;
            int size = ((8 * nLength) < DT_MAX_VAL_SIZE) ? (8 * nLength) : DT_MAX_VAL_SIZE;
            int packet_size = DT_NRMLTP_SIZE + 1 + size;

            val_buff[count++] = (byte)0xFF;                         /* Header */
            val_buff[count++] = (byte)0xFF;                         /* Header */
            val_buff[count++] = (byte)(packet_size);                /* Data Size */
            val_buff[count++] = (byte)((packet_size) >> 8);         /* Data Size */
            val_buff[count++] = (byte)DT_VALTP_ID;                  /* Variable TP */
            val_buff[count++] = (byte)(nStepNumber);                /* nStepNumber */
            val_buff[count++] = (byte)(nStepNumber >> 8);           /* nStepNumber */
            val_buff[count++] = (byte)(nBaseAddress | DT_VARIABLE_BIT);        /* nBaseAddress */
            val_buff[count++] = (byte)(nBaseAddress >> 8);          /* nBaseAddress */
            val_buff[count++] = (byte)(nBaseAddress >> 16);         /* nBaseAddress */
            if(DT_ADD_TIME_INFO == 1){
                DtTimeSetInfo(count);
                count = count + DT_TIME_INFO_SIZE;
            }
            if(DT_ADD_EVENT_INFO == 1){
                DtEventSetInfo(count);
                count = count + DT_EVENT_INFO_SIZE;
            }
            val_buff[count++] = (byte)(size);                       /* size */
            for(int cnt = 0; cnt < nLength && count < packet_size; cnt++){
                val_buff[count++] = (byte)theValue[cnt];            /* value */
                val_buff[count++] = (byte)(theValue[cnt] >> 8);     /* value */
                val_buff[count++] = (byte)(theValue[cnt] >> 16);    /* value */
                val_buff[count++] = (byte)(theValue[cnt] >> 24);    /* value */
                val_buff[count++] = (byte)(theValue[cnt] >> 32);    /* value */
                val_buff[count++] = (byte)(theValue[cnt] >> 40);    /* value */
                val_buff[count++] = (byte)(theValue[cnt] >> 48);    /* value */
                val_buff[count++] = (byte)(theValue[cnt] >> 56);    /* value */
            }
            TP_MemoryOutput(packet_size);
        }
    }
    static public final void DtTestPointValue(int nBaseAddress, int nStepNumber, boolean[] theValue, int nLength){
        if( nLength > 0 ){
            int count = 0;
            int size = ((1 * nLength) < DT_MAX_VAL_SIZE) ? (1 * nLength) : DT_MAX_VAL_SIZE;
            int packet_size = DT_NRMLTP_SIZE + 1 + size;

            val_buff[count++] = (byte)0xFF;                             /* Header */
            val_buff[count++] = (byte)0xFF;                             /* Header */
            val_buff[count++] = (byte)(packet_size);                    /* Data Size */
            val_buff[count++] = (byte)((packet_size) >> 8);             /* Data Size */
            val_buff[count++] = (byte)DT_VALTP_ID;                      /* Variable TP */
            val_buff[count++] = (byte)(nStepNumber);                    /* nStepNumber */
            val_buff[count++] = (byte)(nStepNumber >> 8);               /* nStepNumber */
            val_buff[count++] = (byte)(nBaseAddress | DT_VARIABLE_BIT);            /* nBaseAddress */
            val_buff[count++] = (byte)(nBaseAddress >> 8);              /* nBaseAddress */
            val_buff[count++] = (byte)(nBaseAddress >> 16);             /* nBaseAddress */
            if(DT_ADD_TIME_INFO == 1){
                DtTimeSetInfo(count);
                count = count + DT_TIME_INFO_SIZE;
            }
            if(DT_ADD_EVENT_INFO == 1){
                DtEventSetInfo(count);
                count = count + DT_EVENT_INFO_SIZE;
            }
            val_buff[count++] = (byte)size;                             /* size */
            for(int cnt = 0; cnt < nLength && count < packet_size; cnt++){
                val_buff[count++] = (byte)(theValue[cnt]?0x01:0x00);    /* value */
            }
            TP_MemoryOutput(packet_size);
        }
    }

    /* open socket */
    static public void TpBusEthernetInit() {
        try {
            socket = new DatagramSocket();
            serverInet = InetAddress.getByName(DT_IP_ADDRESS);
            packet = new DatagramPacket(buff, DT_NRMLTP_SIZE, serverInet, DT_PORT);
            val_packet = new DatagramPacket(val_buff, DT_MAX_VALTP_SIZE, serverInet, DT_PORT);
            if(DT_ADD_TIME_INFO == 1){
            	start_time = System.nanoTime();
            }
        }
        catch (Exception e) {
        }
    }

    /* close socket */
    static public void TpBusEthernetFinalize()
    {
        try {
            socket.close();
        }
        catch (Exception e) {
        }
    }

    /* Normal TestPoint Method */
    static public final void TpBusOut(int nAddress, int nData){
        int count = 0;
        if (socket == null) {
            TpBusEthernetInit();
            if(socket == null){
                return;
            }
        }
        buff[count++] = (byte)0xFF;             /* Header */
        buff[count++] = (byte)0xFF;             /* Header */
        buff[count++] = (byte)DT_NRMLTP_SIZE;   /* Data Size */
        buff[count++] = (byte)0x00;             /* Data Size */
        buff[count++] = (byte)DT_NRMLTP_ID;     /* Normal TP */
        buff[count++] = (byte)(nData);          /* nData */
        buff[count++] = (byte)(nData >> 8);     /* nData */
        buff[count++] = (byte)(nAddress);       /* nAddress */
        buff[count++] = (byte)(nAddress >> 8);  /* nAddress */
        buff[count++] = (byte)(nAddress >> 16); /* nAddress */
        if(DT_ADD_TIME_INFO == 1){
            time = TimeUnit.MICROSECONDS.convert( System.nanoTime() - start_time, TimeUnit.NANOSECONDS );
            buff[count++] = (byte)(time);
            buff[count++] = (byte)(time >> 8);
            buff[count++] = (byte)(time >> 16);
            buff[count++] = (byte)(time >> 24);
            buff[count++] = (byte)(time >> 32);
            buff[count++] = (byte)(time >> 40);
            buff[count++] = (byte)(time >> 48);
            buff[count++] = (byte)(time >> 56);
        }
        if(DT_ADD_EVENT_INFO == 1){
            event = (int)Thread.currentThread().getId();
            buff[count++] = (byte)(event);
            buff[count++] = (byte)(event >> 8);
            buff[count++] = (byte)(event >> 16);
            buff[count++] = (byte)(event >> 24);
        }
        try{
            packet.setLength(DT_NRMLTP_SIZE);
            socket.send(packet);
        }
        catch(IOException e){
        }
    }
}

